from torch import mode
from .utils import convertForBatchHardTripletLoss, convertForContrastiveLoss, convertForTripletLoss, convertForSoftmaxLoss
from .batchHardDataset import BatchHardDataset, PKSampler
from .softmaxDataset import SoftmaxDataset
from sentence_transformers import losses , SentenceLabelDataset, SentencesDataset
from torch.utils.data import DataLoader
from .customLosses import SoftmaxLayerLoss



class LossHandler:
    def __init__(self, lossName : str, FAQ, model, batchSize : int):
        supportedLosses = ("batchHardTriplet", "contrastiveLoss", "tripletLoss", "softmaxLayerLoss")
        assert lossName in supportedLosses  , "Error {} loss not supported , supported losses are {}".format(lossName, supportedLosses)

        if(lossName == supportedLosses[0]):
            assert batchSize > 4 and batchSize%4 == 0, "needed for the pk sampler (p ,k > 2)"

            self.trainExamples = convertForBatchHardTripletLoss(FAQ= FAQ)
            self.dataset = BatchHardDataset(examples = self.trainExamples , model = model)
            self.trainLoss = losses.BatchHardTripletLoss(model= model)
            sampler = PKSampler(dataSource= self.dataset, p = 4 , k = batchSize//4)
            self.trainDataLoader =  DataLoader(self.dataset, batch_size= batchSize, sampler= sampler)
        elif(lossName == supportedLosses[1]):
            assert batchSize > 4 and batchSize%4 == 0, "needed for good contrastive loss"
            
            self.trainExamples = convertForContrastiveLoss(FAQ = FAQ)
            self.dataset = SentencesDataset(self.trainExamples, model= model) 
            self.trainLoss = losses.ContrastiveLoss(model = model)
            self.trainDataLoader = DataLoader(self.dataset, shuffle=True, batch_size= batchSize)

        elif(lossName == supportedLosses[2]):
            self.trainExamples = convertForTripletLoss(FAQ = FAQ)
            self.dataset = SentenceLabelDataset(self.trainExamples , model= model)
            self.trainLoss = losses.TripletLoss(model= model , triplet_margin= 1)
            self.trainDataLoader = DataLoader(self.dataset, shuffle=True, batch_size= batchSize)
        elif(lossName == supportedLosses[3]):
            self.trainExamples, numLabels = convertForSoftmaxLoss(FAQ = FAQ)
            self.dataset = SoftmaxDataset(self.trainExamples, model= model)
            self.trainLoss = SoftmaxLayerLoss(model= model, numLabels= numLabels)
            self.trainDataLoader = DataLoader(self.dataset, shuffle= True, batch_size= batchSize)
            

        
        

   


import os
from typing import List, Tuple, Dict

from numpy.testing._private.utils import assert_raises
from .core.FAQ import Answer, FAQ, Question , FAQUnit, FAQOutput
import numpy as np
from sentence_transformers.readers import InputExample
from sentence_transformers import  SentenceTransformer,SentencesDataset
from .core.exceptions import *
from .modelRelated.utils import cosineSim
from .modelRelated.lossHandler import LossHandler
import warnings

class FAQWrapper:
    def __init__(self,id : int ,  FAQ : FAQ):
        self.FAQ = FAQ
        self.id = id
        self.vectors = self.getVectors()


    def getVectors(self) -> np.ndarray:
        faq = self.FAQ.FAQ
        vectors = []
        for unit in faq:
            vector = unit.vectorRep
            if(vector is None):
                raise VectorNotAssignedException()
            vectors.append(vector)
        

        return np.array(vectors)



    def _getClosestQuestions(self,rankedIndices : List[int] ,K : int, topAnswer : str):
        includedSet = set()
        includedSet.add(topAnswer)

        closestQuestions = []
        for ind in rankedIndices:
            currentUnit = self.FAQ.FAQ[ind]
            currentOrignal = currentUnit.orignal.text

            if(currentOrignal not in includedSet):
                includedSet.add(currentOrignal)
                closestQuestions.append(currentOrignal)

            if(len(closestQuestions) == K):
                break

        return closestQuestions

            


        

    def solveForQuery(self,queryVector : np.ndarray, K : int, topSimilar : int = 5)  -> FAQOutput:
        # queryVector has shape (1,emeddingDim)
        if(len(queryVector.shape) == 1):
            queryVector = queryVector.reshape(1,-1)

        assert queryVector.shape[0] == 1 

        cosineScores = cosineSim(queryVector, self.vectors)[0]

        cosineScores = cosineScores.tolist()
        rankedIndices  = [x for x in range(len(cosineScores))]
        rankedIndices.sort(reverse = True, key = lambda x : cosineScores[x])

        maxScore = cosineScores[rankedIndices[0]]
        # Now rankedIndices hold the order of indices with highest to lowest similirity !!!


        competeDict = dict()
        for ind in rankedIndices[:K]:
            # using top K results !!!
            currentlabel = self.FAQ.FAQ[ind].label
            if(currentlabel not in competeDict):
                competeDict[currentlabel] = 0
            competeDict[currentlabel] += cosineScores[ind]

        
        competeList = [(label,score) for label , score in competeDict.items()]
        competeList.sort(key= lambda x : x[1] , reverse= True)

        bestScore = competeList[0][1]
        bestLabel = competeList[0][0]

        bestAnswer = self.FAQ.getAnswerWithLabel(bestLabel)
        bestMatchQuestion = self.FAQ.getQuestionWithLabel(bestLabel)

        return  FAQOutput(faqId= self.id,faqName= self.FAQ.name, answer = bestAnswer,
            question= bestMatchQuestion , score= bestScore,
            similarQuestions=self._getClosestQuestions(rankedIndices,topSimilar,bestMatchQuestion.text) , maxScore= maxScore)
    

        
        




class Bani:
    def __init__(self,FAQs : List[FAQ], modelPath : str = None, assignVectors : bool = True):
        if(modelPath == None):
            modelPath = 'roberta-base-nli-stsb-mean-tokens'
        self.model : SentenceTransformer = self._getModel(modelPath)
        self.FAQs : List[FAQWrapper] = []
        self.idToFAQ : Dict[int,FAQWrapper] = dict()
        self.assignVectors = assignVectors

        self._registerFAQs(FAQs = FAQs)


    def _registerFAQs(self,FAQs : List[FAQ]):
        """
        registers all the faqs given and then extracts vectors , and forms a 
        gobal index and vector to use for combined question answering 
        """

        assert len(FAQs) > 0
        # All FAQs should be Usable !!!
        for faq in FAQs:
            if(faq.isUsable() == False):
                raise ValueError("All faqs passed to chadBot must be Usable !!!! please build FAQ again or load from preexisting one")
        

        for faq in FAQs:
            if(faq.hasVectorsAssigned() == True  and self.assignVectors == False):
                warnings.warn("Vectors already assigned to {} FAQ , if you want to reassign using the current model please clear the vectors using resetAssigned vectors".format(faq.name))
            elif(faq.hasVectorsAssigned() == True and self.assignVectors == True):
                print("OverWritting the vectors of FAQ named {} , it will name save the generated vectors , to do that use the saveFAQ/s feature".format(faq.name))
                faq._assignVectors(model = self.model)
            elif(faq.hasVectorsAssigned() == False and self.assignVectors == True):
                print("Assigning vectors to {} faq , , it will name save the generated vectors , to do that use the saveFAQ/s feature".format(faq.name))
                faq._assignVectors(model = self.model)
            else:
                raise VectorNotAssignedException("assignVectors is False , but the vectors are not stored in faq named {}".format(faq.name))
                
               


        id = 0
        for faq in FAQs:
            newFAQ = FAQWrapper(id,faq)
            self.FAQs.append(newFAQ)
            self.idToFAQ[id] = newFAQ
            id += 1



    def findClosest(self,query : str,  K : int = 3 , topSimilar : int = 5) -> List[FAQOutput]:
        """
        Here we find the closest from each faq and then compare of the 
        top contenders from different faqs are not dangerouusly similar

        """

        competeList : List[FAQOutput] = []
        queryVector = self.model.encode([query])[0].reshape(1,-1)

        for faq in self.FAQs:
            competeList.append(faq.solveForQuery(queryVector=queryVector, K = K, topSimilar= topSimilar))


        competeList.sort(key = lambda x : x.score, reverse= True)
        # competeList now has answer from each faq in the descending order
        return competeList

         
    def findClosestFromFAQ(self,faqId : int, query : str, K : int = 3, topSimilar : int = 5) -> FAQOutput:
        assert faqId in self.idToFAQ , "The id {} not in faqId only {} are available".format(faqId,list(self.idToFAQ.keys()))
        faq = self.idToFAQ[faqId]
        queryVector = self.model.encode([query])[0].reshape(1,-1)
        return faq.solveForQuery(queryVector= queryVector, K = K,topSimilar= topSimilar)


    def _tester(self,faqId : int, questions : List[str],  labels : List[int] , K : int = 3):
        correct = 0
        for question, label in zip(questions,labels):
            answer = self.findClosestFromFAQ(faqId,question)
            if(answer.question.label == label):
                correct += 1

        return correct/len(questions)



    def test(self,faqId : int,testData : List[Tuple[str,str]], K : int = 3) -> float:
        """
        Interface to test any given faq , expects a list of tuples of size 2
        first element is the orignal question and second is the paraphrased version.
        All the orignal question should ideally match the questions in the FAQ
        """
        assert faqId in self.idToFAQ , "The id {} not in faqId only {} are available".format(faqId,list(self.idToFAQ.keys()))
        questions : List[str] = []
        labels : List[str] = []

        nonMatched : List[str] = []
        testFAQ = self.idToFAQ[faqId].FAQ
        for testPoint in testData:
            orignal = testPoint[0]
            reframed = testPoint[1]
            flag = 0
            for unit in testFAQ.FAQ:
                if(unit.orignal.text.strip().lower() == orignal.strip().lower()):
                    questions.append(reframed)
                    labels.append(unit.label)
                    flag = 1
                    break
            if(flag == 0):
                nonMatched.append(orignal)

        if(nonMatched):
            warnings.warn("{} questions in the test set did not match any orignal questions".format(len(nonMatched)))

        print("Running test on {} questions".format(len(questions)))
        return self._tester(faqId= faqId,questions= questions, labels= labels,K= K)


    
    
    
    def train(self,outputPath : str,batchSize = 16, epochs : int = 1, lossName : str = "batchHardTriplet",**kwargs):
        """
        Trains the model using batch hard triplet loss , 
        for the other kwargs take a look at the documentation for sentencetransformers
        """
        os.makedirs(outputPath, exist_ok=True)
        trainingObjectives = [] # training each faq on a different objective
        
        for faq in self.FAQs:
            lossInstance = LossHandler(lossName= lossName, model= self.model, FAQ = faq.FAQ, batchSize= batchSize)
            trainDataloader = lossInstance.trainDataLoader
            trainLoss = lossInstance.trainLoss
            trainingObjectives.append((trainDataloader, trainLoss))

        self.model.fit(train_objectives=  trainingObjectives, warmup_steps= 100,epochs= epochs, save_best_model= False,output_path= outputPath, **kwargs)
        self.saveModel(outputPath)
        self.model = SentenceTransformer(outputPath)
        for faq in self.FAQs:
            print("Assigning vectors from the trained model to FAQ {}".format(faq.FAQ.name))
            faq.FAQ._assignVectors(self.model)



    def getFAQWithId(self, id : int) -> FAQ:
        if(id not in self.idToFAQ):
            raise KeyError("FAQ with id {} does not exist".format(id))

        return self.idToFAQ[id].FAQ

    def saveModel(self, path):
        os.makedirs(path, exist_ok=True)
        self.model.save(path)


    def _getModel(self, path):
        return SentenceTransformer(path)

    def saveFAQs(self, rootDirPath : str):
        for faq in self.FAQs:
            coreFaQ = faq.FAQ
            coreFaQ.save(rootDirPath)
    def saveFAQ(self, id : int, rootDirPath : str):
        assert id in self.idToFAQ
        self.idToFAQ[id].FAQ.save(rootDirPath)

    
        


if __name__ == "__main__":
    print("HEllo")







        





       





    









        
        
        


    
    


    




    

    














from sentence_transformers.readers import InputExample
from ..core.FAQ import FAQ, FAQUnit
from ..core.exceptions import *
from typing import Dict, List
import random
import numpy as np


def cosineSim(v,V):
    """
    computes cosine sim between v,V
    where v and V are  2D np matrices (n,E) (N,E)
    output is of the shape (n,N)
    """

    n1 = np.linalg.norm(v, axis = -1)
    n2 = np.linalg.norm(V, axis = -1)
    dot = np.expand_dims(v,1)*np.expand_dims(V,0)
    # shape (n,N,E)
    dot = dot.sum(axis = -1)
    ans = dot/n1.reshape(-1,1)
    ans = ans/n2.reshape(1,-1)
    return ans



def minLabelChecker(lossName : str , FAQ : FAQ, minSameLabels : int):
    faq = FAQ.FAQ

    counter = dict()
    for unit in faq:
        label = unit.label

        if(label not in counter):
            counter[label] = 0
        counter[label] += 1

    for L,count in counter.items():
        if(count < minSameLabels):
            raise TrainDataInvalid("Attempted to use {} loss but label {} has only {} exmaples !!! At least {} required".format(lossName,L,count,minSameLabels))




def convertForBatchHardTripletLoss(FAQ : FAQ, minSameLabels = 2) -> List[InputExample]:
    """
    Must check if the data has atleast minSameLabels(atleast 2) examples for each label !!!
    """
    assert minSameLabels >= 2
    minLabelChecker(lossName= "batchHardTriplet",FAQ = FAQ,minSameLabels= minSameLabels)
    faq = FAQ.FAQ    
    # now converting
    inputs = []
    for unit in faq:
        inputs.append(InputExample(texts = [unit.question.text] , label= unit.label))

    
    return inputs



convertForTripletLoss =  convertForBatchHardTripletLoss



def convertForContrastiveLoss(FAQ : FAQ , minSameLabels : int = 2, numIters : int = 1):
    """
    Gets data for training on a cosine similirity loss, each sentence must have atleaast one paraphrase !!!
    numIters is number of times we pick up negative examples
    """
    assert minSameLabels >= 2 and numIters >= 1
    minLabelChecker(lossName= "ContrastiveLoss",FAQ = FAQ,minSameLabels= minSameLabels)

    faq = FAQ.FAQ

    inputs = []

    labelToQuestions : Dict[int, List[str]] = dict()


    for unit in faq:
        label = unit.label
        if(label not in labelToQuestions):
            labelToQuestions[label] = []
        labelToQuestions[label].append(unit.question.text)

    for label in list(labelToQuestions.keys()):
        # positives
        positives = labelToQuestions[label]
        
        for i in range(len(positives)):
            for j in range(i+ 1, len(positives)):
                s1, s2 = positives[i], positives[j]
                inputs.append(InputExample(texts= [s1,s2], label= 1))



    for _ in range(numIters):
        labelList = list(labelToQuestions.keys())
        for i in range(len(labelList)):
            # negatives
            for j in range(i+1,len(labelList)):
                label1, label2 = labelList[i],labelList[j]
                if(label2 == label1):
                    continue


                s1 = random.choice(labelToQuestions[label1])
                s2 = random.choice(labelToQuestions[label2])

                inputs.append(InputExample(texts= [s1,s2], label= 0))

        

    return inputs





def convertForSoftmaxLoss(FAQ : FAQ):
    """
    This loss adds an softmax layer on top of the sentence transformer
    """

    # making faq labels to loss label mapping

    labelTolabel : Dict[int,int] = dict()
    labelToQuestions : Dict[int,List[str]] = dict()   # Contains mapping from the loss labels to questions 
    faq = FAQ.FAQ
    index = 0
    for unit in faq:
        label = unit.label
        if(label not in labelTolabel):
            labelTolabel[label] = index
            index += 1

        lossLabel = labelTolabel[label]
        if(lossLabel not in labelToQuestions):
            labelToQuestions[lossLabel] = []
        labelToQuestions[lossLabel].append(unit.question.text)



    inputs = []

    for label , questions in labelToQuestions.items():
        for question in questions:
            inputs.append(InputExample(texts= [question], label= label))
    
    random.shuffle(inputs)
    return inputs , len(labelToQuestions)




    


from numpy.core.fromnumeric import size
from sentence_transformers import InputExample
from typing import List, Union , Dict, Tuple, Any
import random
import numpy as np
from torch.utils.data.dataset import Dataset
from torch.utils.data.sampler import Sampler
import torch


class BatchHardDataset(Dataset):
    def __init__(self, examples : List[InputExample],  model):
        for example in examples:
            assert len(example.texts) == 1
        
        self.model = model
        self.examples : List[InputExample] = examples

        self._build_dict()
        for example in self.examples:
            example.texts_tokenized = self.tokenizeExample(example=example)


    def _build_dict(self):
        self.labelToIndices : Dict[int , List[int]] = dict()

        for i,example in enumerate(self.examples):
            label = example.label
            if(label not in self.labelToIndices):
                self.labelToIndices[label] = []
            
            self.labelToIndices[label].append(i)

    
    def tokenizeExample(self, example : InputExample):
        if example.texts_tokenized is not None:
            return example.texts_tokenized

        return [self.model.tokenize(text) for text in example.texts]
    
    
    def __getitem__(self, idx : int) -> Tuple[List[Any], torch.Tensor]:
        return self.examples[idx].texts_tokenized , torch.tensor(self.examples[idx].label)


    def __len__(self) -> int:
        return len(self.examples)



class PKSampler(Sampler):
    def __init__(self, dataSource : BatchHardDataset, p=4, k=4):
        super().__init__(dataSource)
        assert p > 2 and k > 2 , "batch hard condition enforced, increase batch size"
        self.p = p
        self.k = k
        self.dataSource = dataSource

    def __iter__(self):
        pkCount = len(self) // (self.p * self.k)
        for _ in range(pkCount):
            labels = np.random.choice(list(self.dataSource.labelToIndices.keys()), self.p, replace=False)
            for l in labels:
                indices = self.dataSource.labelToIndices[l]
                replace = True if len(indices) < self.k else False

                for i in np.random.choice(indices, self.k, replace=replace):
                    yield i

    def __len__(self):
        pk = self.p * self.k
        samples = ((len(self.dataSource) - 1) // pk + 1) * pk
        return samples


        

        



import torch
from torch import nn, Tensor
from typing import Union, Tuple, List, Iterable, Dict
import torch.nn.functional as F
from enum import Enum

class SoftmaxLayerLoss(nn.Module):
    def __init__(self, model , numLabels):
        super().__init__()
        self.model = model
        self.numlabels = numLabels
        self.modelOutDim = self.model.get_sentence_embedding_dimension()
        self.classifier = nn.Linear(self.modelOutDim,numLabels, bias= False)


    def forward(self, sentence_features: Iterable[Dict[str, torch.Tensor]], labels: torch.Tensor):
        features = [self.model(sentence_feature)['sentence_embedding'] for sentence_feature in sentence_features][0]
        output = self.classifier(features)
        loss_fct = nn.CrossEntropyLoss()
        loss = loss_fct(output, labels.view(-1))
        return loss


from torch.utils.data.dataset import Dataset
import torch
from sentence_transformers import InputExample
from typing import List, Union , Dict, Tuple, Any
import random
import numpy as np




class SoftmaxDataset(Dataset):
    def __init__(self, examples : List[InputExample], model):
        self.model = model
        self.examples = examples
        for example in self.examples:
            example.texts_tokenized = self.tokenizeExample(example=example)





    def tokenizeExample(self, example : InputExample):
        if example.texts_tokenized is not None:
            return example.texts_tokenized

        return [self.model.tokenize(text) for text in example.texts]


    def __len__(self) -> int:
        return len(self.examples)
    

    def __getitem__(self, idx: int) -> Any:
        return self.examples[idx].texts_tokenized , torch.tensor(self.examples[idx].label)

